import csv
import json
from pathlib import Path
import re
from typing import NamedTuple


class SearchStimulus(NamedTuple):
    """class that represents a visual search stimulus"""
    stimulus: str
    set_size: int
    target_condition: str
    img_num: int
    root_output_dir: str
    img_file: str
    meta_file: str

    @classmethod
    def cast_row(cls, row):
        """Convert string values in given dictionary
        to corresponding SearchStimulus field type.
        """
        return {field: cls._field_types[field](value)
                for field, value in row.items()}


FIELDNAMES = list(SearchStimulus._fields)


def make_csv(rows, csv_filename):
    """utility function to make csv

    Parameters
    ----------
    rows : list
        values that will go into csv, with each element
        corresponding to a field defined by the FIELDNAMES
        constant declared in this module (searchstims.utils.FIELDNAMES).
    csv_filename : str
        name of csv file that will be created

    Notes
    -----
    Each row in a .csv file corresponds to an image generated by searchstims.
    Fields in csv files are:
        stimulus : str
            name of visual search stimulus, e.g. 'RVvGV'.
        set_size : int
            visual search set size, i.e. total number of targets and distractors
            in the stimulus.
        target_condition : str
            one of {'present', 'absent'}. Whether stimulus contains target.
        img_num : int
            a unique index assigned to the image by the main loop that creates them.
            Makes it possible to distinguish multiple images that would otherwise have
            the same filename (e.g. because they both have set size of 8 and target
            present, but the items are in different locations).
        root_output_dir : str
            location specified by user where all images are saved
        img_file : str
            relative path to image from root_output_dir.
            So the following produces the absolute path:
            >>> Path(root_output_dir).joinpath(img_file)
        meta_file : str
            relative path to .json metadata file associated with the image.
            The metadata file contains the indices of the centers of items
            in the image, i.e. their location. If the images were generated
            using a grid (instead of randomly placing items) then the metadata
            will also contain a string representation of where items are in the
            grid. This string representation can be used to find e.g. all stimuli
            of set size 8 where the target was on the left side of the image.
    """
    with open(csv_filename, 'w') as f:
        writer = csv.DictWriter(f, FIELDNAMES)
        writer.writeheader()
        rowdicts = [
            dict(zip(FIELDNAMES, row))
            for row in rows
        ]
        writer.writerows(rowdicts)


def json_to_csv(json_filename, root_output_dir):
    """convert .json file output by earlier versions to a csv.
    Also creates individual .meta.json files for each individual .png file.

    Parameters
    ----------
    json_filename : str
        name of .json file to convert to .csv file
    root_output_dir : str
        absolute path to directory that is root of directories + images
        created by searchstims

    Notes
    -----
    The format of the .json files is described here for future reference
    (and so the programmer can figure out how to go from complicated nested
     dictionaries in .json to a flat "long-form" .csv):

    The .json file is a serialized Python dictionary of dictionaries
    with the following key, field pairs:

    {set size: {
        'present': [
            {'filename': str,
             'grid_as_char': list,
            ]

    Keys at the top level are set size, the total number of targets and distractors, e.g.,
    {1, 2, ..., 8}. Each set size key has as its value another dictionary,
    whose keys are 'present' and 'absent', referring to the visual search target.
    Each 'present' and 'absent' key has as its value a list of Python dictionaries;
    each dictionary in the list has info about the actual visual search stimulus image
    that it corresponds to:
        filename: str
            actual visual search stimulus filename
        grid_as_char: list
            of list of str. Representation of stimulus as a grid of
            cells
        target_indices: list
            of two-element lists, the x and y co-ordinates for the
            center of the targets (or indices if you load the image into
            an array).
        distractor_indices: list
            of two-element lists, the x and y co-ordinates for the
            center of the distractors (or indices if you load the image into
            an array).

    Here is an excerpt from such a file:
        {'1': {'absent': [{'distractor_indices': [[203, 65]],
            'filename': '/home/user/output/1/absent/redvert_v_greenvert_set_size_1_target_absent_0.png',
            'grid_as_char':
                [['', '', '', '', ''],
                 ['', '', '', '', 'd'],
                 ['', '', '', '', ''],
                 ['', '', '', '', ''],
                 ['', '', '', '', '']],
            'target_indices': []},
           {'distractor_indices': [[111, 21]],
            'filename': '/home/user/output/1/absent/redvert_v_greenvert_set_size_1_target_absent_1.png',
            ...
         '2': {'absent': [{'distractor_indices': [[68, 22], [65, 204]],
            'filename': '/home/user/output/2/absent/redvert_v_greenvert_set_size_2_target_absent_0.png',
            'grid_as_char': [['', 'd', '', '', ''],
            ...
    """
    with open(json_filename) as fp:
        metadict = json.load(fp)

    root_output_dir = Path(root_output_dir)
    if not root_output_dir.is_dir():
        raise NotADirectoryError(
            "value for `root_output_dir` not recognized as a directory: "
            f"{root_output_dir}"
        )

    root_output_dir = root_output_dir.absolute()

    rows = []
    for stimulus, stim_dict in metadict.items():
        for set_size, set_size_dict in stim_dict.items():
            for target_condition, list_meta_dict in set_size_dict.items():
                for meta_dict in list_meta_dict:
                    img_file = meta_dict['filename']
                    nums = re.findall('\d+', img_file)
                    img_num = nums[-1]
                    meta_file = img_file.replace('.png', '.meta.json')
                    row = [stimulus,
                           set_size,
                           target_condition,
                           img_num,
                           root_output_dir,
                           img_file,
                           meta_file]
                    rows.append(row)

                    # make .meta.json file too
                    meta_dict['img_file'] = meta_dict.pop('filename')
                    meta_file = Path(root_output_dir).joinpath(meta_file)
                    with open(meta_file, 'w') as fp:
                        json.dump(meta_dict, fp)

    # get just name, in case 'csv' appears somewhere in path like a parent directory
    csv_name = Path(json_filename).name
    csv_name = csv_name.replace('json', 'csv')
    # now assemble path to csv file using json_filename parent plus csv_name
    csv_filename = str(
        Path(json_filename).parent.joinpath(csv_name)
    )
    make_csv(rows, csv_filename)



import os
import configparser
from distutils.util import strtobool
import ast

from .classes import Config, GeneralConfig, RVvGVConfig, Two_v_Five_Config

SECTION_CONFIG_ATTRIB_MAP = {
    'general': 'general',
    'RVvGV': 'RVvGV',
    '2_v_5': 'Two_v_Five',
}


SECTION_CLASS_MAP = {
    'general': GeneralConfig,
    'RVvGV': RVvGVConfig,
    '2_v_5': Two_v_Five_Config,
}


this_file_dir = os.path.dirname(__file__)

CONFIG_TYPES = configparser.ConfigParser()
CONFIG_TYPES.read(os.path.join(this_file_dir, 'types.ini'))
VALID_SECTIONS = CONFIG_TYPES.sections()

DEFAULT_CONFIG = configparser.ConfigParser()
DEFAULT_CONFIG.read(os.path.join(this_file_dir, 'default.ini'))


def parse(config_file=None):
    """read config.ini file with config parser,
    returns namedtuple ConfigTuple with
    sections and options as attributes.

    Parameters
    ----------
    config_file : str
        Path to a config.ini file. If None, the default configuration is returned.
        Default is None.

    Returns
    -------
    config : searchstims.config.classes.Config
        instance of Config class that represents sections of config.ini file,
        with attributes general, number, and rectangle, each of which is an
        instance of a class that represents those sections with attributes being
        the options in those sections and the corresponding values set by the
        user.
    """
    if not os.path.isfile(config_file):
        raise FileNotFoundError(f'config_file {config_file} not found')

    config = configparser.ConfigParser()
    config.read(config_file)

    if not config.has_section('general'):
        raise ValueError(
            f"'general' section required in configuration file but not found in {config_file}"
        )

    # validate sections and set default values from default.ini
    for section in config.sections():
        # is section name valid?
        if section not in VALID_SECTIONS:
            raise ValueError(f'invalid section name: {section}.\n'
                             f'Valid section names are: {VALID_SECTIONS}')
        # if so, check for undeclared options and if not declared, set to defaults from default.ini
        for option in DEFAULT_CONFIG.options(section):
            if not config.has_option(section, option):
                config[section][option] = DEFAULT_CONFIG[section][option]

    if config.has_option(section='general',
                         option='output_dir'):
        output_dir = config['general']['output_dir']
        config['general']['output_dir'] = os.path.expanduser(output_dir)

    sections = [key for key in list(config.keys()) if key != 'DEFAULT']
    typed_config_dict = {}
    for section in sections:
        section_keys = list(config[section].keys())
        section_values = list(config[section].values())
        section_dict = {}
        for key, val in zip(section_keys, section_values):
            if val == 'None':
                typed_val = None
            else:
                val_type = CONFIG_TYPES[section][key]
                if val_type == 'int':
                    typed_val = int(val)
                elif val_type == 'float':
                    typed_val = float(val)
                elif val_type == 'bool':
                    typed_val = bool(strtobool(val))
                elif val_type == 'list' or val_type == 'tuple':
                    typed_val = ast.literal_eval(val)
                elif val_type == 'str':
                    typed_val = val
            section_dict[key] = typed_val
        typed_config_dict[section] = section_dict

    # if user didn't declare some stim section, set to None
    for section in VALID_SECTIONS:
        if section not in typed_config_dict:
            typed_config_dict[section] = None

    config_classes_dict = {}
    for section in typed_config_dict.keys():
        if typed_config_dict[section] is not None:
            section_attrib = SECTION_CONFIG_ATTRIB_MAP[section]
            section_class = SECTION_CLASS_MAP[section]
            config_classes_dict[section_attrib] = section_class(**typed_config_dict[section])

    config_obj = Config(**config_classes_dict)

    return config_obj



from pathlib import Path
import random

import pygame
from pygame.rect import Rect

from .abstract_stim_maker import AbstractStimMaker
from .abstract_stim_maker import colors_dict

THIS_FILE_DIR = Path(__file__).parent


class xoStimMaker(AbstractStimMaker):

    forcedsquare_path = str(THIS_FILE_DIR.joinpath('..', 'ttf', 'forced_square.ttf'))

    """Make visual search stimuli where target is 'x' and distractors are 'x's and 'o's.
    Distractors can be same or different colors.
    """
    def __init__(self,
                 target_x_color='blue',
                 distractor_x_color='red',
                 distractor_o_color='red',
                 **kwargs):
        super().__init__(target_color=target_x_color,
                         distractor_color=distractor_x_color,
                         **kwargs)

        self.target_x_color = target_x_color
        self.distractor_x_color = distractor_x_color
        self.distractor_o_color = distractor_o_color

    def _make_stim(self,
                   xx_to_use_ctr,
                   yy_to_use_ctr,
                   target_inds,
                   cells_to_use,
                   display_surface):
        """helper function used by make_stim that sub-classes can override if they need to do something more
        complicated when making the visual search stimulus

        uses xx_to_use_ctr and yy_to_use_ctr to create item_bbox Rects for each item in the visual search stimulus.
        If the item's index is in the list of target_inds then it is a target and the target color is used.

        Each item is drawn while looping through (xx_to_use_ctr, yy_to_use_ctr) by calling self.draw_item
        """
        target_indices = []
        distractor_indices = []
        if self.grid_size:
            grid_as_char = [''] * self.num_cells
        else:
            grid_as_char = None

        # divide distractors up into x and o
        set_size = len(xx_to_use_ctr)
        num_distractors = set_size - len(target_inds)
        num_distractor_o = set_size // 2
        if len(target_inds) == 0:
            num_distractor_x = set_size // 2 - len(target_inds)
        else:
            num_distractor_x = set_size // 2

        if num_distractors % 2 == 1:  # e.g., if odd set size and target absent
            diff = num_distractors - (num_distractor_o + num_distractor_x)
            for _ in range(diff):
                if random.uniform(0, 1) > 0.5:
                    num_distractor_o += 1
                else:
                    num_distractor_x += 1

        distractor_letters = list('x' * num_distractor_x + 'o' * num_distractor_o)
        random.shuffle(distractor_letters)

        for item, (center_x, center_y) in enumerate(zip(xx_to_use_ctr, yy_to_use_ctr)):
            # notice we are now using PyGame order of sizes, (width, height)
            item_bbox_tuple = (0, 0) + (self.item_bbox_size[1], self.item_bbox_size[0])
            item_bbox = Rect(item_bbox_tuple)
            center = (int(center_x), int(center_y))
            item_bbox.center = center

            if item in target_inds:
                is_target = True
                color = self.target_x_color
                target_indices.append(center)
                if self.grid_size:
                    grid_as_char[cells_to_use[item]] = 't'
            else:
                is_target = False
                distractor_letter = distractor_letters.pop()
                if distractor_letter == 'x':
                    color = self.distractor_x_color
                elif distractor_letter == 'o':
                    color = self.distractor_o_color
                distractor_indices.append(center)
                if self.grid_size:
                    grid_as_char[cells_to_use[item]] = distractor_letter

            if type(color) == str:
                color = colors_dict[color]

            font_obj = pygame.font.Font(self.forcedsquare_path, 64)
            if is_target:
                text_surface_obj = font_obj.render('x', True, color)
            else:
                text_surface_obj = font_obj.render(distractor_letter, True, color)
            text_surface_obj = pygame.transform.scale(text_surface_obj,
                                                      (self.item_bbox_size[1], self.item_bbox_size[0]))

            self.draw_item(display_surface=display_surface,
                           item_bbox=item_bbox,
                           to_blit=text_surface_obj)

        return grid_as_char, target_indices, distractor_indices

    def draw_item(self, display_surface, item_bbox, to_blit):
        """Returns target or distractor"""
        display_surface.blit(to_blit, item_bbox)



from .abstract_stim_maker import AbstractStimMaker
from .RVvGVStimMaker import RVvGVStimMaker
from .RVvRHGVStimMaker import RVvRHGVStimMaker
from .Two_v_Five_StimMaker import Two_v_Five_StimMaker
from .TStimMaker import TStimMaker
from .TLStimMaker import TLStimMaker
from .xoStimMaker import xoStimMaker



import argparse
import os

from .config import parse
from .make import make


from .stim_makers import RVvGVStimMaker, Two_v_Five_StimMaker
from .config.parse import SECTION_CONFIG_ATTRIB_MAP

VALID_STIM_SECTIONS = [
    'RVvGV',
    '2_v_5'
]


COMMON_ARGS = [
        'image_size',
        'border_size',
        'grid_size',
        'min_center_dist',
        'item_bbox_size',
        'jitter',
]


def _get_common_args_from_stim_or_general(stim_config, general_config):
    """helper function that gets arguments from general config if they are not
    specified in the config.ini section corresponding to a visual search stimulus
    
    Parameters
    ----------
    stim_config : RectangleConfig or NumberConfig
        instance of class that represent config.ini section corresponding to a visual search stimulus
    general_config : GeneralConfig
        instance of class that represents config.ini section with general configuration options

    Returns
    -------
    common_args : tuple
        of the arguments listed in searchstims.main.COMMON_ARGS
    """
    common_args = []
    for this_attr in COMMON_ARGS:
        this_arg = getattr(stim_config, this_attr)
        if this_arg is None:
            this_arg = getattr(general_config, this_attr)
        common_args.append(this_arg)
    return tuple(common_args)


def _get_stim_dict(config):
    """for each section in config_obj, if that section is a valid stimulius name,
    then generate a stim_maker for it

    Parameters
    ----------
    config : searchstims.config.classes.Config
        instance of Config returned by parse_config
        after parsing a config.ini file
    """
    general_config = config.general

    stim_dict = {}
    for section in VALID_STIM_SECTIONS:
        section_attr_name = SECTION_CONFIG_ATTRIB_MAP[section]
        if getattr(config, section_attr_name) is not None:  # because config.section will be None if not in config.ini
            stim_config = getattr(config, section_attr_name)
            (image_size,
             border_size,
             grid_size,
             min_center_dist,
             item_bbox_size,
             jitter) = _get_common_args_from_stim_or_general(stim_config, general_config)

            if section == 'RVvGV':
                stim_maker = RVvGVStimMaker(target_color=stim_config.target_color,
                                            distractor_color=stim_config.distractor_color,
                                            window_size=image_size,
                                            border_size=border_size,
                                            grid_size=grid_size,
                                            min_center_dist=min_center_dist,
                                            item_bbox_size=item_bbox_size,
                                            jitter=jitter
                                            )

            elif section == '2_v_5':
                stim_maker = Two_v_Five_StimMaker(target_color=stim_config.target_color,
                                                  distractor_color=stim_config.distractor_color,
                                                  window_size=image_size,
                                                  border_size=border_size,
                                                  grid_size=grid_size,
                                                  min_center_dist=min_center_dist,
                                                  item_bbox_size=item_bbox_size,
                                                  jitter=jitter,
                                                  target_number=stim_config.target_number,
                                                  distractor_number=stim_config.distractor_number
                                                  )
            stim_dict[section] = stim_maker

    return stim_dict


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('configfile',
                        type=str,
                        help=('filename of config.ini file, e.g.:\n'
                              '$ searchstims ./basic_config.ini\n'
                              'For an example config.ini file, see: '
                              'https://github.com/NickleDave/searchstims'))
    args = parser.parse_args()
    config_file = args.configfile
    if not os.path.isfile(config_file):
        raise FileNotFoundError("Config file {} not found".format(config_file))
    config = parse(config_file)
    stim_dict = _get_stim_dict(config)
    make(root_output_dir=config.general.output_dir,
         stim_dict=stim_dict,
         json_filename=config.general.json_filename,
         num_target_present=config.general.num_target_present,
         num_target_absent=config.general.num_target_absent,
         set_sizes=config.general.set_sizes)


if __name__ == '__main__':
    main()



from itertools import combinations, product
import json
from math import ceil
from pathlib import Path
import random

import numpy as np
import pygame

from .stim_makers import AbstractStimMaker
from .utils import make_csv


def _generate_xx_and_yy(set_size,
                        num_imgs,
                        stim_maker):
    """helper function that computes x,y co-ordinates for items in visual search stimulus

    ensures that there are no repeated images in dataset

    finds number of combinations of cells given set size of stimulus and grid size specified for it
    """
    # get all combinations of cells (combination because order doesn't matter, just which cells get used)
    # a cell combination is an unordered set of k cells from a grid with a total of n cells
    # e.g. if there are 25 cells in a 5x5 grid and you want all combinations k=1, then the
    # cell_combs will be [(0,), (1,), (2,), ... (24,)] (representing each as a tuple)
    # and all combinations k=2 will be [(0,1), (0,2), (0,3), ... (1,2), (1,3), ... (23, 24)]
    # (there are no repeats; once we draw a cell we don't replace it since we just put one item in each cell)
    cell_combs = list(combinations(iterable=range(stim_maker.num_cells), r=set_size))

    # if there are less combinations then there are number of images, we need to make sure we make jitter
    # unique so we don't get any repeat images
    if len(cell_combs) < num_imgs:
        # num_repeat: maximum number of times we might use any given cell combination
        num_repeat = ceil(num_imgs / len(cell_combs))
        make_jitter_unique = True
    else:
        # don't need to repeat any cell combinations; let's just sample without replacement
        all_cells_to_use = random.sample(population=cell_combs, k=num_imgs)
        num_repeat = 0
        make_jitter_unique = False

    # if jitter > 0, compute 'jitter_range'
    # (vector of possible jitter amounts within maximum jitter from which to draw randomly)
    if stim_maker.jitter > 0:
        jitter_high = stim_maker.jitter // 2
        jitter_low = -jitter_high
        if stim_maker.jitter % 2 == 0:  # if even
            # have to account for zero at center of jitter range
            # (otherwise range would be jitter + 1)
            # (Not a problem when doing floor division on odd #s)
            coin_flip = np.random.choice([0, 1])
            if coin_flip == 0:
                jitter_low += 1
            elif coin_flip == 1:
                jitter_high -= 1
        jitter_range = range(jitter_low, jitter_high + 1)
        # get cartesian product of jitter range of length 2, i.e. all x-y co-ordinates
        # here we want cartesian product because order does matter, jitter of (1,0) != (0, 1)
        # and because we want repeats, e.g. (2, 2)
        jitter_coords = list(product(jitter_range, repeat=2))

        if make_jitter_unique:
            # get each unique pairing of possible cell combinations and possible x, y jitters
            if len(cell_combs) * len(jitter_coords) < num_imgs:
                raise ValueError('cannot generate unique x and y co-ordinates for items in number of images specified; '
                                 f'the product of the number of cell combinations {len(all_cells_to_use)} and the '
                                 f'possible jitter added {len(jitter_coords)} is {len(cell_combs) * len(jitter_coords)}'
                                 f', but the number of images to generate is {num_imgs}')
            else:
                cell_and_jitter = []
                for this_cell_comb in cell_combs:
                    jitter_sample = random.sample(population=jitter_coords, k=num_repeat)
                    this_cell_comb_with_jitter = [(this_cell_comb, jitter_coord_tup)
                                                  for jitter_coord_tup in jitter_sample]
                    cell_and_jitter.extend(this_cell_comb_with_jitter)
                diff = len(cell_and_jitter) - num_imgs
                # remove extras randomly instead of removing all from the last cell_comb
                inds_to_remove = random.sample(range(len(cell_and_jitter)), k=diff)
                inds_to_remove.sort(reverse=True)
                for ind in inds_to_remove:
                    cell_and_jitter.pop(ind)
                all_cells_to_use = [cj_tup[0] for cj_tup in cell_and_jitter]
        else:
            jitter_rand = random.choices(jitter_coords, k=len(all_cells_to_use))
            cell_and_jitter = zip(all_cells_to_use, jitter_rand)
    else:  # if jitter == 0
        jitter_none = [None] * len(all_cells_to_use)
        cell_and_jitter = zip(all_cells_to_use, jitter_none)

    ###########################################################################
    # notice: below we always refer to y before x, because shapes are         #
    # specified in order of (height, width). So size[0] = y and size[1] = x   #
    ###########################################################################
    all_yy_to_use_ctr = []
    all_xx_to_use_ctr = []
    for cells_to_use, jitter_to_add in cell_and_jitter:
        # need to cast to list and then array because converting a list with asarray returns a 1-dimensional
        # array, and indexing with this one-dimensional array returns another 1-d array. Using just a tuple
        # or an array made from a tuple will just return a single element when the tuple has only one element,
        # and this will raise an error when that one element is passed to stim_maker instead of a 1-d, 1 element
        # array
        cells_to_use = np.asarray(list(cells_to_use))
        yy_to_use = stim_maker.yy[cells_to_use]
        xx_to_use = stim_maker.xx[cells_to_use]

        # find centers of cells we're going to use
        yy_to_use_ctr = (yy_to_use * stim_maker.cell_height) - stim_maker.cell_y_center
        xx_to_use_ctr = (xx_to_use * stim_maker.cell_width) - stim_maker.cell_x_center

        if stim_maker.border_size:
            yy_to_use_ctr += round(stim_maker.border_size[0] / 2)
            xx_to_use_ctr += round(stim_maker.border_size[1] / 2)

        if jitter_to_add:
            yy_to_use_ctr += jitter_to_add[0]
            xx_to_use_ctr += jitter_to_add[1]

        all_yy_to_use_ctr.append(yy_to_use_ctr)
        all_xx_to_use_ctr.append(xx_to_use_ctr)

    return all_cells_to_use, all_xx_to_use_ctr, all_yy_to_use_ctr


def make(root_output_dir,
         stim_dict,
         csv_filename,
         num_target_present,
         num_target_absent,
         set_sizes):
    """make visual search stimuli given an output directory and a set of StimMaker classes

    Parameters
    ----------
    root_output_dir : str, Path
        directory in which output should be saved
    stim_dict : dict
        key, value pairs where the key is the visual search stimulus name and the 'value' is
        an instance of a StimMaker
    csv_filename : str
        name for .csv file that will be saved containing metadata about generated set of images (see Notes below).
    num_target_present : int, list
        number of visual search stimuli to generate with target present.
        If int, the number of stimuli generated for each set size will be num_target_present // len(set_sizes).
        List should be same length as set_sizes with an int value for each set size, the number of 'target present'
        images to make for that set size.
    num_target_absent : int, list
        number of visual search stimuli to generate with target absent.
        If int, the number of stimuli generated for each set size will be num_target_absent // len(set_sizes).
        List should be same length as set_sizes with an int value for each set size, the number of 'target absent'
        images to make for that set size.
    set_sizes : list
        of int, e.g. [1, 2, 4, 8]. If num_target_present and num_target_absent are ints, not lists, then
        the number of stimuli generated for each set size will be num_target_present(absent) // len(set_size).
        E.g., 4800 / 4 = 1200 images per set size. If num_target_present/absent are lists then each should
        be the same length as set_sizes and the value at the index corresponding to each set size determines the
        number of stimuli generated for that set size. E.g. if num_target_present = [1000, 2000, 4000] and
        set_sizes = [1, 2, 4] then there will be 1000 stimuli with set size 1, 2000 with set size 2, and 4000
        with set size 4.

    Returns
    -------
    None

    Notes
    -----
    This function saves all the stimuli to root_output_dir, and saves information about
    stimuli in a .csv output file. The .csv file has the following fields:
        stimulus : str
            name of visual search stimulus, e.g. 'RVvGV'.
        set_size : int
            visual search set size, i.e. total number of targets and distractors
            in the stimulus.
        target_condition : str
            one of {'present', 'absent'}. Whether stimulus contains target.
        img_num : int
            a unique index assigned to the image by the main loop that creates them.
            Makes it possible to distinguish multiple images that would otherwise have
            the same filename (e.g. because they both have set size of 8 and target
            present, but the items are in different locations).
        root_output_dir : str
            location specified by user where all images are saved
        img_file : str
            relative path to image from root_output_dir.
            So the following produces the absolute path:
            >>> Path(root_output_dir).joinpath(img_file)
        meta_file : str
            relative path to .json metadata file associated with the image.
            The metadata file contains the indices of the centers of items
            in the image, i.e. their location. If the images were generated
            using a grid (instead of randomly placing items) then the metadata
            will also contain a string representation of where items are in the
            grid. This string representation can be used to find e.g. all stimuli
            of set size 8 where the target was on the left side of the image.

    The .json metadata files are Python dictionaries with the following key/value pairs:
        img_filename : str
            name of image file that metadata is associated with.
            Included in metadata so user does not have to extract it from filename.
        target_indices : list
            co-ordinates of targets, i.e. indices in array representing image.
            A Numpy array converted to a list.
        distractor_indices : list
            co-ordinates of distractors, i.e. indices in array representing image
            A Numpy array converted to a list.
        grid_as_char : str
            Representation of stimulus as characters. This is only added when the stimulus
            is generated as a grid where the items can appear within cells on the grid.
            If the stimulus is generated with another method, e.g. randomly placing items,
            then the value for this key will be None.

    Here is an example dictionary from a metadata file:
    {'img_filename': 'RVvGV/1/absent/redvert_v_greenvert_set_size_1_target_absent_0.png',
     'target_indices': [],
     'distractor_indices': [[203, 65]],
     'grid_as_char':
            [['', '', '', '', ''],
             ['', '', '', '', 'd'],
             ['', '', '', '', ''],
             ['', '', '', '', ''],
             ['', '', '', '', '']],
    }
    """
    for stim_name, stim_maker in stim_dict.items():
        if type(stim_name) != str:
            raise TypeError(
                f'all keys in stim_dict must be strings but found key of type {type(stim_name)}'
            )
        if not issubclass(type(stim_maker), AbstractStimMaker):
            raise TypeError(
                f'stim_maker not recognized as a subclass of AbstractStimMaker, type was {stim_maker}'
            )

    if type(num_target_present) not in (int, list):
        raise TypeError(
            f'num_target_present should be int or list but type was: {type(num_target_present)}'
        )

    if type(num_target_present) is list:
        if len(num_target_present) != len(set_sizes):
            raise ValueError(
                'num_target_present must be same length as set_sizes'
            )

        if not all([type(num) is int for num in num_target_present]):
            raise ValueError(
                'all values in num_target_present should be int'
            )

    if type(num_target_absent) not in (int, list):
        raise TypeError(
            f'num_target_present should be int or list but type was: {type(num_target_absent)}'
        )

    if type(num_target_absent) is list:
        if len(num_target_absent) != len(set_sizes):
            raise ValueError(
                'num_target_absent must be same length as set_sizes'
            )

        if not all([type(num) is int for num in num_target_absent]):
            raise ValueError(
                'all values in num_target_absent should be int'
            )

    if type(root_output_dir) == str:
        root_output_dir = Path(root_output_dir)

    if not root_output_dir.is_dir():
        root_output_dir.mkdir(parents=True)

    root_output_dir = root_output_dir.absolute()

    # for csv
    rows = []

    for stimulus, stim_maker in stim_dict.items():
        stimulus_output_dir = root_output_dir.joinpath(stimulus)
        if not stimulus_output_dir.is_dir():
            stimulus_output_dir.mkdir()

        # if num_target_present/absent are int, make into list
        # so we can zip them with set_sizes in main loop
        if type(num_target_present) is int:
            num_target_present = num_target_present // len(set_sizes)
            num_target_present = [num_target_present for _ in range(len(set_sizes))]

        if type(num_target_absent) is int:
            num_target_absent = num_target_absent // len(set_sizes)
            num_target_absent = [num_target_absent for _ in range(len(set_sizes))]

        for set_size, num_imgs_present, num_imgs_absent in zip(
                set_sizes, num_target_present, num_target_absent):

            set_size_dir = stimulus_output_dir.joinpath(str(set_size))
            if not set_size_dir.is_dir():
                set_size_dir.mkdir()

            for target_condition in ('present', 'absent'):
                if target_condition == 'present':
                    img_nums = list(range(num_imgs_present))
                    num_target = 1
                elif target_condition == 'absent':
                    img_nums = list(range(num_imgs_absent))
                    num_target = 0

                target_condition_dir = set_size_dir.joinpath(target_condition)
                if not target_condition_dir.is_dir():
                    target_condition_dir.mkdir()

                def _make_stim(img_num, cells_to_use=None,
                               xx_to_use_ctr=None, yy_to_use_ctr=None):
                    """helper function to make and save individual stim

                    Define as a nested function so we can avoid repeating ourselves below
                    """
                    rect_tuple = stim_maker.make_stim(set_size=set_size,
                                                      num_target=num_target,
                                                      cells_to_use=cells_to_use,
                                                      xx_to_use_ctr=xx_to_use_ctr,
                                                      yy_to_use_ctr=yy_to_use_ctr)

                    filename = (
                        f'{stimulus}_set_size_{set_size}_target_{target_condition}'
                        f'_{img_num}.png'
                    )
                    abs_path_filename = target_condition_dir.joinpath(filename)
                    pygame.image.save(rect_tuple.display_surface,
                                      str(abs_path_filename))
                    # use relative path for name of file in csv
                    # so it won't break anything if we move the whole directory of images
                    # we can just change 'root_output_dir' instead
                    img_file = Path(stimulus).joinpath(str(set_size),
                                                       target_condition,
                                                       filename)
                    meta_file = Path(
                        str(abs_path_filename).replace('.png', '.meta.json')
                    )
                    meta_dict = {
                        'img_file': str(img_file),
                        'target_indices': rect_tuple.target_indices,
                        'distractor_indices': rect_tuple.distractor_indices,
                        'grid_as_char': rect_tuple.grid_as_char,
                    }
                    with open(meta_file, 'w') as fp:
                        json.dump(meta_dict, fp)

                    row = (stimulus,
                           set_size,
                           target_condition,
                           img_num,
                           root_output_dir,
                           img_file,
                           meta_file)
                    rows.append(row)

                if stim_maker.grid_size is None:
                    for img_num in img_nums:
                        _make_stim(img_num)
                else:
                    (all_cells_to_use,
                     all_xx_to_use_ctr,
                     all_yy_to_use_ctr) = _generate_xx_and_yy(set_size=set_size,
                                                              num_imgs=len(img_nums),
                                                              stim_maker=stim_maker)

                    for img_num, cells_to_use, xx_to_use_ctr, yy_to_use_ctr in zip(img_nums,
                                                                                   all_cells_to_use,
                                                                                   all_xx_to_use_ctr,
                                                                                   all_yy_to_use_ctr):
                        _make_stim(img_num, cells_to_use, xx_to_use_ctr, yy_to_use_ctr)

    csv_filename = root_output_dir.joinpath(csv_filename)
    make_csv(rows, csv_filename)



from collections import namedtuple

import pygame
from pygame.locals import *
import numpy as np
from scipy.spatial.distance import pdist

pygame.init()

# set up colors
colors_dict = {
    'black': (0, 0, 0),
    'white': (255, 255, 255),
    'red255green51': (255, 51, 0),
    'green255red51': (51, 255, 0),
    'red255green102': (255, 102, 0),
    'green255red102': (102, 255, 0),
    'red255green153': (255, 153, 0),
    'green255red153': (153, 255, 0),
    'red255green204': (255, 204, 0),
    'green255red204': (204, 255, 0),
    'red': (255, 0, 0),
    'green': (0, 255, 0),
    'blue': (0, 0, 255),
}

MAX_DRAWS_INNER = 1000
MAX_DRAWS_OUTER = 100


def validate_color(color):
    if type(color) not in (str, tuple):
        raise TypeError(
            f'color must be a string or a three-element tuple corresponding to an RGB color but was: {type(color)}'
        )

    if type(color) == tuple:
        if len(color) != 3:
            raise ValueError(
                f'color tuple must be a 3 element RGB color; number of elements was {len(color)}'
            )
        if not all([type(el) == int for el in color]):
            raise ValueError(
                f'all elements in color tuple should be of type int, color values were: {color}'
            )


class AbstractStimMaker:
    """parent class for all StimMaker classes

    Attributes
    ----------
    target_color : str
        {'red', 'green', 'white', 'blue'}. Default is 'red'.
    distractor_color : str
        {'red', 'green', 'white', 'blue'}. Default is 'green'.
    window_size : tuple
        of length two, representing (height, width) of window in pixels.
    border_size : tuple
        of length two, representing (height, width) of border, distance
        from edge of window within which items should not be displayed.
        Default is None.
    grid_size : tuple
        of length two, representing the number of (rows, columns)
        in the grid on which target and distractors will be located.
    min_center_dist : int
        Minimum distance between center point of items. Default is None, in
        which case any distance is permitted. Only used if grid_size is None
        and items are placed randomly instead of on a grid.
    item_bbox_size : tuple
        shape of "bounding box" that contains items to be plotted,
        (height, width) in pixels. Default is (30, 30).
    jitter : int
        number of pixels to jitter each 'item' in the 'set'
        that will be plotted on the grid. Default is 5.
        Adding jitter is helpful for creating stimuli with the
        same set size but slightly different placements, e.g. for
        augmenting data to train a learning algorithm and encourage
        invariant representations.
    num_cells : int
        rows * columns in grid
    grid_size_pixels : tuple
        (width, height). Size of grid in pixels. Either equal to window size, or if
        border size is not None, equal to window size minus border size.
    yy : numpy.ndarray
       y-axis co-ordinates of cells, e.g. [0, 1, 2, 3, 4] for an y-axis with 5 cells.
    xx : numpy.ndarray
        x-axis co-ordinates of cells, e.g. [0, 1, 2, 3, 4] for an x-axis with 5 cells.
    cell_height : int
        height of a cell in grid, in pixels
    cell_y_center : int
        co-ordinate of center of y axis of a cell. Add to co-ordinate of cell corner to
        get the center point of cell within the entire window.
    cell_width : int
        width of a cell in grid, in pixels
    cell_x_center
        co-ordinate of center of x axis of a cell. Add to co-ordinate of cell corner to
        get the center point of cell within the entire window.
    """

    RectTuple = namedtuple('RectTuple', ['display_surface',
                                         'grid_as_char',
                                         'target_indices',
                                         'distractor_indices'])

    def __init__(self,
                 target_color='red',
                 distractor_color='green',
                 background_color='black',
                 window_size=(227, 227),
                 border_size=None,
                 grid_size=(5, 5),
                 min_center_dist=None,
                 item_bbox_size=(30, 30),
                 jitter=5):
        """__init__ function for Stim Makers

        Parameters
        ----------
        target_color : str, tuple
            One of {'red', 'green', 'blue', 'white', 'black'}. Default is 'red'.
        distractor_color : str, tuple
            One of {'red', 'green', 'blue', 'white', 'black'}. Default is 'green'.
        background_color : str, tuple
            One of {'red', 'green', 'blue', 'white', 'black'}. Default is 'black'.
        window_size : tuple
            of length two, representing (height, width) of window in pixels.
        border_size : tuple
            of length two, representing (height, width) of border, distance
            from edge of window within which items should not be displayed.
            Default is None.
        grid_size : tuple
            of length two, representing the number of (rows, columns)
            in the grid on which target and distractors will be located.
        min_center_dist : int
            Minimum distance between center point of items. Default is None, in
            which case any distance is permitted. Only used if grid_size is None
            and items are placed randomly instead of on a grid.
        item_bbox_size : tuple
            shape of "bounding box" that contains items to be plotted,
            (height, width) in pixels. Default is (30, 30).
        jitter : int
            number of pixels to jitter each 'item' in the 'set'
            that will be plotted on the grid. Default is 5.
            Adding jitter is helpful for creating stimuli with the
            same set size but slightly different placements, e.g. for
            augmenting data to train a learning algorithm and encourage
            invariant representations.
        """
        if grid_size is not None:
            if not all([type(grid_size_el) == int for grid_size_el in grid_size]):
                raise ValueError('values for grid size must be positive integers')

            if not all([grid_size_el > 0 for grid_size_el in grid_size]):
                raise ValueError('values for grid size must be positive integers')

        for color in [target_color, distractor_color, background_color]:
            validate_color(color)

        self.target_color = target_color
        self.distractor_color = distractor_color
        self.background_color = background_color
        self.grid_size = grid_size
        self.min_center_dist = min_center_dist
        self.window_size = window_size
        self.border_size = border_size
        self.item_bbox_size = item_bbox_size
        self.jitter = jitter

        if self.grid_size:
            self.num_cells = self.grid_size[0] * self.grid_size[1]

            if self.border_size is None:
                self.grid_size_pixels = self.window_size
            else:
                self.grid_size_pixels = (self.window_size[0] - self.border_size[0],
                                         self.window_size[1] - self.border_size[1])

            # make grid, randomly select which cells in grid to use.
            grid_y = np.arange(1, self.grid_size[0] + 1)
            grid_x = np.arange(1, self.grid_size[1] + 1)
            yy, xx = np.meshgrid(grid_y, grid_x)
            self.yy = yy.ravel()
            self.xx = xx.ravel()

            # find centers of cells we're going to use
            self.cell_height = round(self.grid_size_pixels[0] / self.grid_size[0])
            self.cell_y_center = round((self.grid_size_pixels[0] / self.grid_size[0]) / 2)

            self.cell_width = round(self.grid_size_pixels[1] / self.grid_size[1])
            self.cell_x_center = round((self.grid_size_pixels[1] / self.grid_size[1]) / 2)

    def draw_item(self, display_surface, item_bbox, color=None, to_blit=None):
        """draw item for visual search stimulus

        Parameters
        ----------
        display_surface : pygame.Surface
        item_bbox : pygame.rect
            item bounding box. The actual item is drawn
            *within* this bounding box.
        color : tuple
            3-item tuple, i.e. RGB color.
        to_blit : pygame.image
            image to blit onto item_bbox
        """
        raise NotImplementedError

    def _make_stim(self,
                   xx_to_use_ctr,
                   yy_to_use_ctr,
                   target_inds,
                   cells_to_use,
                   display_surface):
        """helper function used by make_stim that sub-classes can override if they need to do something more
        complicated when making the visual search stimulus

        uses xx_to_use_ctr and yy_to_use_ctr to create item_bbox Rects for each item in the visual search stimulus.
        If the item's index is in the list of target_inds then it is a target and the target color is used.

        Each item is drawn while looping through (xx_to_use_ctr, yy_to_use_ctr) by calling self.draw_item
        """
        target_indices = []
        distractor_indices = []
        if self.grid_size:
            grid_as_char = [''] * self.num_cells
        else:
            grid_as_char = None

        for item, (center_x, center_y) in enumerate(zip(xx_to_use_ctr, yy_to_use_ctr)):
            # notice we are now using PyGame order of sizes, (width, height)
            item_bbox_tuple = (0, 0) + (self.item_bbox_size[1], self.item_bbox_size[0])
            item_bbox = Rect(item_bbox_tuple)
            center = (int(center_x), int(center_y))
            item_bbox.center = center

            if item in target_inds:
                color = self.target_color
                target_indices.append(center)
                if self.grid_size:
                    grid_as_char[cells_to_use[item]] = 't'
            else:
                color = self.distractor_color
                distractor_indices.append(center)
                if self.grid_size:
                    grid_as_char[cells_to_use[item]] = 'd'

            if type(color) == str:
                color = colors_dict[color]

            self.draw_item(display_surface=display_surface,
                           item_bbox=item_bbox,
                           color=color)

        return grid_as_char, target_indices, distractor_indices

    def make_stim(self,
                  set_size=8,
                  num_target=1,
                  cells_to_use=None,
                  xx_to_use_ctr=None,
                  yy_to_use_ctr=None
                  ):
        """make visual search stimuli

        Parameters
        ----------
        set_size : int
            Set size, equal to number of targets + distractors.
            Default is 8.
        num_target : int
            Number of targets. Default is 1.
        cells_to_use : list
            list of cells in grid to use. Length must equal set_size.
            Default is None, in which case cells are drawn randomly from
            grid defined for this stim maker. The centers of the cells of
            the grid are accessible as stim_maker.xx and stim_maker.yy
        xx_to_use_ctr : numpy.ndarray
            co-ordinates of center of item bounding boxes on x axis.
            One-dimensional vector; number of elements must equal set_size.
            Default is None. If None and cells_to_use is None, cells will
            be drawn at random and the center of those cells used.
        yy_to_use_ctr : numpy.ndarray
            co-ordinates of center of item bounding boxes on y axis.
            One-dimensional vector; number of elements must equal set_size.
            Default is None. If None and cells_to_use is None, cells will
            be drawn at random and the center of those cells used.

        Returns
        -------
        display_surface : pygame.Surface
            with visual search stimuli plotted on it
        """
        if type(set_size) != int:
            raise TypeError('set size must be an integer')

        if not set_size > 0:
            raise ValueError('set size must be greater than zero')

        if type(num_target) != int:
            raise TypeError('number of targets must be an integer')

        if not num_target >= 0:
            raise ValueError('number of targets must be greater than or equal to zero')

        if num_target > set_size:
            raise ValueError('number of targets cannot be greater than set size')

        if cells_to_use:
            if len(cells_to_use) != set_size:
                raise ValueError(f'Number of elements in cells_to_use must equal set_size.\n'
                                 f'cells_to_use has length {len(cells_to_use)} but set_size is {set_size}.')

        if xx_to_use_ctr is not None:
            if yy_to_use_ctr is None:
                raise ValueError('must pass an argument for yy_to_use_ctr when passing xx_to_use_ctr')
            if xx_to_use_ctr.ndim != 1:
                raise ValueError('xx_to_use_ctr should be a one-dimensional vector but has shape: '
                                 f'{xx_to_use_ctr.shape}')
            if xx_to_use_ctr.shape[0] != set_size:
                raise ValueError('Number of elements in xx_to_use_ctr must equal set size.'
                                 f'xx_to_use_ctr.shape is {xx_to_use_ctr.shape} and set size is {set_size}')

        if yy_to_use_ctr is not None:
            if xx_to_use_ctr is None:
                raise ValueError('must pass an argument for xx_to_use_ctr when passing yy_to_use_ctr')
            if yy_to_use_ctr.ndim != 1:
                raise ValueError('yy_to_use_ctr should be a one-dimensional vector but has shape: '
                                 f'{yy_to_use_ctr.shape}')
            if yy_to_use_ctr.shape[0] != set_size:
                raise ValueError('Number of elements in yy_to_use_ctr must equal set size.'
                                 f'yy_to_use_ctr.shape is {yy_to_use_ctr.shape} and set size is {set_size}')

        if self.grid_size is not None:
            total_grid_elements = self.grid_size[0] * self.grid_size[1]
            if set_size > total_grid_elements:
                raise ValueError('set size {} cannot be greater than number of '
                                 'elements in grid, {}'
                                 .format(set_size, total_grid_elements))

        if type(self.jitter) != int:
            raise TypeError('value for jitter must be an integer')

        ###########################################################################
        # notice: below we always refer to y before x, because shapes are         #
        # specified in order of (height, width). So size[0] = y and size[1] = x   #
        ###########################################################################
        if cells_to_use is not None or (xx_to_use_ctr is None and yy_to_use_ctr is None):
            if self.grid_size:
                if cells_to_use is None:
                    cells_to_use = sorted(np.random.choice(np.arange(self.num_cells),
                                                           size=set_size,
                                                           replace=False))

                    yy_to_use = self.yy[cells_to_use]
                    xx_to_use = self.xx[cells_to_use]

                    # find centers of cells we're going to use
                    yy_to_use_ctr = (yy_to_use * self.cell_height) - self.cell_y_center

                    xx_to_use_ctr = (xx_to_use * self.cell_width) - self.cell_x_center

                    if self.border_size:
                        yy_to_use_ctr += round(self.border_size[0] / 2)
                        xx_to_use_ctr += round(self.border_size[1] / 2)

                    # add jitter to those points
                    jitter_high = self.jitter // 2
                    jitter_low = -jitter_high
                    if self.jitter % 2 == 0:  # if even
                        # have to account for zero at center of jitter range
                        # (otherwise range would be jitter + 1)
                        # (Not a problem when doing floor division on odd #s)
                        coin_flip = np.random.choice([0, 1])
                        if coin_flip == 0:
                            jitter_low += 1
                        elif coin_flip == 1:
                            jitter_high -= 1
                    jitter_range = np.arange(jitter_low, jitter_high + 1)

                    y_jitter = np.random.choice(jitter_range, size=yy_to_use.size)
                    yy_to_use_ctr += y_jitter
                    x_jitter = np.random.choice(jitter_range, size=xx_to_use.size)
                    xx_to_use_ctr += x_jitter

            else:  # if self.grid_size is None
                if self.border_size:
                    yy = np.arange(self.border_size[0] + (self.item_bbox_size[0] / 2),
                                   self.window_size[0] - (self.border_size[0] + (self.item_bbox_size[0] / 2) + 1),
                                   dtype=int)
                    xx = np.arange(self.border_size[1] + (self.item_bbox_size[1] / 2),
                                   self.window_size[1] - (self.border_size[1] + (self.item_bbox_size[1] / 2) + 1),
                                   dtype=int)
                else:
                    yy = np.arange(self.item_bbox_size[0] / 2,
                                   self.window_size[0] - (self.item_bbox_size[0] / 2))
                    xx = np.arange(self.item_bbox_size[1] / 2,
                                   self.window_size[1] - (self.item_bbox_size[1] / 2))

                # draw center points at random
                dists_are_good = False
                less_than_set_size = True
                draws_outer = 0
                while dists_are_good is False:
                    draws_outer += 1
                    if draws_outer > MAX_DRAWS_OUTER:
                        raise ValueError('could not find suitable set of co-ordinates for set')

                    draws_inner = 0
                    coords_list = []

                    while less_than_set_size is True:
                        yy_to_use_ctr = np.random.choice(yy)
                        xx_to_use_ctr = np.random.choice(xx)
                        coord = [xx_to_use_ctr, yy_to_use_ctr]
                        draws_inner += 1
                        if draws_inner > MAX_DRAWS_INNER:
                            coords_list = []
                            draws_inner = 0

                        if len(coords_list) == 0:
                            coords_list.append(coord)

                            if set_size == 1:
                                # then we're done actually
                                less_than_set_size = False
                                dists_are_good = True
                                # make into an array of size (1,) to prevent crash when we index into them below
                                yy_to_use_ctr = np.asarray([yy_to_use_ctr])
                                xx_to_use_ctr = np.asarray([xx_to_use_ctr])
                        else:
                            coords_list_tmp = list(coords_list)
                            coords_list_tmp.append(coord)
                            coords_list_tmp = np.stack(coords_list_tmp)
                            dists = pdist(coords_list_tmp)

                            if np.any(dists < self.min_center_dist):
                                continue
                            else:
                                coords_list = coords_list_tmp.tolist()

                                if len(coords_list) < set_size:
                                    continue
                                else:
                                    coords = np.asarray(coords_list)
                                    yy_to_use_ctr = coords[:, 1]
                                    xx_to_use_ctr = coords[:, 0]
                                    dists_are_good = True
                                    less_than_set_size = False

        # set up window
        display_surface = pygame.display.set_mode((self.window_size[1], self.window_size[0]),
                                                  0,
                                                  32)

        # draw on surface object
        display_surface.fill(colors_dict[self.background_color])
        target_inds = np.random.choice(np.arange(set_size),
                                       size=num_target).tolist()

        # call helper function that actually makes search stimulus
        # (added so that sub-classes can override just that function if they need to)
        (grid_as_char,
         target_indices,
         distractor_indices) = self._make_stim(xx_to_use_ctr,
                                               yy_to_use_ctr,
                                               target_inds,
                                               cells_to_use,
                                               display_surface)

        if self.grid_size:
            grid_as_char = np.asarray(grid_as_char).reshape(self.grid_size[0], self.grid_size[1]).tolist()

        pygame.display.update()
        return self.RectTuple(display_surface=display_surface,
                              grid_as_char=grid_as_char,
                              target_indices=target_indices,
                              distractor_indices=distractor_indices)





from pathlib import Path

import pygame
from pygame.rect import Rect

from .abstract_stim_maker import AbstractStimMaker
from .abstract_stim_maker import colors_dict

THIS_FILE_DIR = Path(__file__).parent


class Two_v_Five_StimMaker(AbstractStimMaker):
    """Make visual search stimuli where the target is a digital 2
    and the distractors are digital 5s."""
    forcedsquare_path = str(THIS_FILE_DIR.joinpath('..', 'ttf', 'forced_square.ttf'))

    def __init__(self,
                 target_number=2,
                 distractor_number=5,
                 **kwargs):
        """
        Other Parameters
        ----------------
        target_number : int
            Number that is used as target.
            one of {2, 5}. Default is 2.
        distractor_number : int
            Number that is used as a distractor.
            one of {2, 5}. Default is 5.
        """
        super().__init__(**kwargs)
        self.target_number = str(target_number)
        self.distractor_number = str(distractor_number)

    def draw_item(self, display_surface, item_bbox, to_blit):
        """Returns target or distractor"""
        display_surface.blit(to_blit, item_bbox)

    def _make_stim(self,
                   xx_to_use_ctr,
                   yy_to_use_ctr,
                   target_inds,
                   cells_to_use,
                   display_surface):
        """helper function used by make_stim
        that sub-classes can override if they need to do something more
        complicated when making the visual search stimulus"""
        target_indices = []
        distractor_indices = []
        if self.grid_size:
            grid_as_char = [''] * self.num_cells
        else:
            grid_as_char = None

        for item, (center_x, center_y) in enumerate(zip(xx_to_use_ctr, yy_to_use_ctr)):
            # notice we are now using PyGame order of sizes, (width, height)
            item_bbox_tuple = (0, 0) + (self.item_bbox_size[1], self.item_bbox_size[0])
            item_bbox = Rect(item_bbox_tuple)
            center = (int(center_x), int(center_y))
            item_bbox.center = center

            if item in target_inds:
                is_target = True
                color = self.target_color
                target_indices.append(center)
                if self.grid_size:
                    grid_as_char[cells_to_use[item]] = 't'
            else:
                is_target = False
                color = self.distractor_color
                distractor_indices.append(center)
                if self.grid_size:
                    grid_as_char[cells_to_use[item]] = 'd'

            if type(color) == str:
                color = colors_dict[color]

            font_obj = pygame.font.Font(self.forcedsquare_path, 64)
            if is_target:
                text_surface_obj = font_obj.render(self.target_number,
                                                   True, color)
            else:
                text_surface_obj = font_obj.render(self.distractor_number,
                                                   True, color)
            text_surface_obj = pygame.transform.scale(text_surface_obj,
                                                      (self.item_bbox_size[1],
                                                       self.item_bbox_size[0]))

            self.draw_item(display_surface=display_surface,
                           item_bbox=item_bbox,
                           to_blit=text_surface_obj)

        return grid_as_char, target_indices, distractor_indices



import pygame

from .abstract_stim_maker import AbstractStimMaker


class RVvGVStimMaker(AbstractStimMaker):
    """Make visual search stimuli with vertical rectangles
    where target is red and distractors are green."""
    def draw_item(self, display_surface, item_bbox, color):
        """Draws a vertical rectangle that is 1/3 the width of the item bounding box.

        Parameters
        ----------
        display_surface : pygame.Surface
            instance of Surface on which to draw item
        item_bbox : pygame.Rect
            instance of Rect that represents 'bounding box' within which
            item should be drawn.
        is_target : bool
            if True, item to draw is target and target color should be used

        Returns
        -------
        None
        """
        rect_to_draw = item_bbox  # yes, they point to the same object

        width = rect_to_draw.width
        width = width // 3
        rect_to_draw.width = width
        rect_to_draw.left = rect_to_draw.left + width

        pygame.draw.rect(display_surface, color, rect_to_draw)



from pathlib import Path

import pygame
from pygame.rect import Rect

from .abstract_stim_maker import AbstractStimMaker
from .abstract_stim_maker import colors_dict

THIS_FILE_DIR = Path(__file__).parent


class TStimMaker(AbstractStimMaker):

    forcedsquare_path = str(THIS_FILE_DIR.joinpath('..', 'ttf', 'forced_square.ttf'))

    """Make visual search stimuli with T shapes, where target is rotated 90 degrees."""
    def __init__(self,
                 target_color='white',
                 distractor_color='white',
                 target_rotation=90,
                 **kwargs):
        if type(target_rotation) != int or not(0 < target_rotation < 360):
            raise ValueError('target_rotation must be an integer between 1 and 359')

        super().__init__(target_color=target_color,
                         distractor_color=distractor_color,
                         **kwargs)

        self.target_rotation = target_rotation

    def _make_stim(self,
                   xx_to_use_ctr,
                   yy_to_use_ctr,
                   target_inds,
                   cells_to_use,
                   display_surface):
        """helper function used by make_stim that sub-classes can override if they need to do something more
        complicated when making the visual search stimulus

        uses xx_to_use_ctr and yy_to_use_ctr to create item_bbox Rects for each item in the visual search stimulus.
        If the item's index is in the list of target_inds then it is a target and the target color is used.

        Each item is drawn while looping through (xx_to_use_ctr, yy_to_use_ctr) by calling self.draw_item
        """
        target_indices = []
        distractor_indices = []
        if self.grid_size:
            grid_as_char = [''] * self.num_cells
        else:
            grid_as_char = None

        for item, (center_x, center_y) in enumerate(zip(xx_to_use_ctr, yy_to_use_ctr)):
            # notice we are now using PyGame order of sizes, (width, height)
            item_bbox_tuple = (0, 0) + (self.item_bbox_size[1], self.item_bbox_size[0])
            item_bbox = Rect(item_bbox_tuple)
            center = (int(center_x), int(center_y))
            item_bbox.center = center

            if item in target_inds:
                is_target = True
                color = self.target_color
                target_indices.append(center)
                if self.grid_size:
                    grid_as_char[cells_to_use[item]] = 't'
            else:
                is_target = False
                color = self.distractor_color
                distractor_indices.append(center)
                if self.grid_size:
                    grid_as_char[cells_to_use[item]] = 'd'

            if type(color) == str:
                color = colors_dict[color]

            font_obj = pygame.font.Font(self.forcedsquare_path, 64)
            text_surface_obj = font_obj.render('T', True, color)
            text_surface_obj = pygame.transform.scale(text_surface_obj,
                                                      (self.item_bbox_size[1], self.item_bbox_size[0]))
            if is_target:
                # rotate AFTER scaling (so same aspect ratio as un-rotated T)
                text_surface_obj = pygame.transform.rotate(text_surface_obj, self.target_rotation)

            self.draw_item(display_surface=display_surface,
                           item_bbox=item_bbox,
                           to_blit=text_surface_obj)

        return grid_as_char, target_indices, distractor_indices

    def draw_item(self, display_surface, item_bbox, to_blit):
        """Returns target or distractor"""
        display_surface.blit(to_blit, item_bbox)



""" classes to represent sections of config.ini file """
import attr
from attr.validators import instance_of, optional
from attr import converters


def check_len_is_two(instance, attribute, value):
    if len(value) != 2:
        raise ValueError(f"{attribute.name} tuple for {instance.name} should be two elements, got {value}")


@attr.s
class GeneralConfig:
    """represents [GENERAL] section of config.ini file

    Attributes
    ----------
    output_dir : str
        path to directory where images and .json file created by running searchstims
        should be saved
    csv_filename : str
        name of .csv file that will be created with information about images created
    enforce_unique : bool
        if True, ensures that each stimulus is unique by drawing without replacement all
        item locations *before* generating any of the stimuli

    The remaining attributes, if declared as options and assigned values in the [GENERAL] section
    of a config.ini file, will be used for all stimuli *unless* the same options are declared in
    a section for a specific stimulus, in which case those values override the values assigned to
    the attributes in the [GENERAL[ section (and thus the GeneralConfig instance that represents it).

    num_target_present : int
        number of visual search stimuli to generate with target present.
    num_target_absent : int
        number of visual search stimuli to generate with target absent
    set_sizes : list
        of int, "set sizes" that should be generated
    item_bbox_size : tuple
        two element tuple, (height, width). The size of the
        "bounding box" that contains items (target + distractors) in the visual search stimulus.
    image_size : tuple
        two element tuple, (height, width). This will be the size of an input
        to the neural network architecture that you're training.
    grid_size : tuple
        two element tuple, (rows, columns). Represents the "grid" that the
        visuals search stimulus is divided into, where each cell in that
        grid can contain an item (either the target or a distractor). The
        total number of cells will be rows * columns.
    border_size : tuple
        two element tuple, (height, width). The size of the border between
        the actual end of the image and the grid of cells within the image
        that will contain the items (target + distractors).
        Useful if you are worried about edge effects.
    min_center_dist : int
        minimum distance to maintain between the center of items.
        Useful if you are worried about crowding effects.
    jitter : int
        maximum value of jitter applied to center points of items.
    """
    output_dir = attr.ib(validator=instance_of(str))
    csv_filename = attr.ib(validator=instance_of(str))
    num_target_present = attr.ib(converter=converters.optional(int))
    num_target_absent = attr.ib(converter=converters.optional(int))
    set_sizes = attr.ib(validator=optional(instance_of(list)))
    enforce_unique = attr.ib(validator=optional(instance_of(bool)), default=True)
    item_bbox_size = attr.ib(validator=optional([instance_of(tuple), check_len_is_two]),
                                 default=None)
    image_size = attr.ib(validator=optional([instance_of(tuple), check_len_is_two]),
                         default=None)
    grid_size = attr.ib(validator=optional([instance_of(tuple), check_len_is_two]),
                        default=None)
    border_size = attr.ib(validator=optional([instance_of(tuple), check_len_is_two]),
                          default=None)
    min_center_dist = attr.ib(validator=optional(instance_of(int)), default=None)
    jitter = attr.ib(validator=optional(instance_of(int)), default=None)


@attr.s
class RVvGVConfig:
    """represent [RVvGV] section of config.ini file

    Attributes
    ----------
    item_bbox_size : tuple
        two element tuple, (height, width). The size of the
        "bounding box" that contains items (target + distractors) in the visual search stimulus.
    image_size : tuple
        two element tuple, (height, width). This will be the size of an input
        to the neural network architecture that you're training.
    grid_size : tuple
        two element tuple, (rows, columns). Represents the "grid" that the
        visuals search stimulus is divided into, where each cell in that
        grid can contain an item (either the target or a distractor). The
        total number of cells will be rows * columns.
    border_size : tuple
        two element tuple, (height, width). The size of the border between
        the actual end of the image and the grid of cells within the image
        that will contain the items (target + distractors).
        Optional; default is None. Useful if you are worried about edge effects.
    min_center_dist : int
        minimum distance to maintain between the center of items.
        Useful if you are worried about crowding effects.
        Optional; default is None.
    jitter : int
        maximum value of jitter applied to center points of items. Default is 5.
    rectangle_size : tuple
        two element tuple, (height, width). The size of the rectangle plotted
        inside the item bounding box. Default is (10, 30).
    target_color : str
        color of target. For RectangleConfig, default is 'red'.
    distractor_color : str
        color of target. For RectangleConfig, default is 'green'.
    """
    item_bbox_size = attr.ib(validator=optional([instance_of(tuple), check_len_is_two]),
                                 default=(30, 30))
    image_size = attr.ib(validator=optional([instance_of(tuple), check_len_is_two]),
                         default=(227, 227))
    grid_size = attr.ib(validator=optional([instance_of(tuple), check_len_is_two]),
                        default=(5,5))
    border_size = attr.ib(validator=optional([instance_of(tuple), check_len_is_two]),
                          default=None)
    min_center_dist = attr.ib(validator=optional(instance_of(int)), default=None)
    jitter = attr.ib(validator=optional(instance_of(int)), default=5)
    target_color = attr.ib(validator=instance_of(str), default='red')
    distractor_color = attr.ib(validator=instance_of(str), default='green')


def number_validator(instance, attribute, value):
    if value not in (2, 5):
        raise ValueError(f'{attribute} must be either 2 or 5, not {value}')
    if (
                (attribute == 'target_number' and value == instance.distractor_number)
            or
                (attribute == 'distractor_number' and value == instance.target_number)
    ):
            raise ValueError(f"target number should not equal distractor number but both are {value}")


@attr.s
class Two_v_Five_Config:
    """represent [2_v_5] section of config.ini file

    Attributes
    ----------
    item_bbox_size : tuple
        two element tuple, (height, width). The size of the
        "bounding box" that contains items (target + distractors) in the visual search stimulus.
        For NumberConfig this must be the size of the .png images that contain the
        number shapes, (30, 30).
    image_size : tuple
        two element tuple, (height, width). This will be the size of an input
        to the neural network architecture that you're training.
    grid_size : tuple
        two element tuple, (rows, columns). Represents the "grid" that the
        visuals search stimulus is divided into, where each cell in that
        grid can contain an item (either the target or a distractor). The
        total number of cells will be rows * columns.
    border_size : tuple
        two element tuple, (height, width). The size of the border between
        the actual end of the image and the grid of cells within the image
        that will contain the items (target + distractors).
        Optional; default is None. Useful if you are worried about edge effects.
    min_center_dist : int
        minimum distance to maintain between the center of items.
        Useful if you are worried about crowding effects.
        Optional; default is None.
    jitter : int
    target_color : int
        color of target. For NumberConfig, default is 'white'.
    distractor_color : int
        color of target. For NumberConfig, default is 'white'.
    target_number : int
    distractor_number : int
    """
    item_bbox_size = attr.ib(validator=optional([instance_of(tuple), check_len_is_two]),
                                 default=(30, 30))
    image_size = attr.ib(validator=optional([instance_of(tuple), check_len_is_two]),
                         default=(227, 227))
    grid_size = attr.ib(validator=optional([instance_of(tuple), check_len_is_two]),
                        default=(5,5))
    border_size = attr.ib(validator=optional([instance_of(tuple), check_len_is_two]),
                          default=None)
    min_center_dist = attr.ib(validator=optional(instance_of(int)), default=None)
    jitter = attr.ib(validator=optional(instance_of(int)), default=5)
    target_color = attr.ib(validator=instance_of(str), default='white')
    distractor_color = attr.ib(validator=instance_of(str), default='white')
    target_number = attr.ib(validator=[instance_of(int), number_validator], default=2)
    distractor_number = attr.ib(validator=[instance_of(int), number_validator], default=5)


@attr.s
class Config:
    """class to represent all sections of config.ini file

    Attributes
    ----------
    general: TrainConfig
        represents [TRAIN] section
    RVvGV: RVvVGConfig
        represents [RVvVG] section
    number: NumberConfig
        represents [2_v_5] section
    """
    general = attr.ib(GeneralConfig)
    RVvGV = attr.ib(validator=optional(instance_of(RVvGVConfig)), default=None)
    Two_v_Five = attr.ib(validator=optional(instance_of(Two_v_Five_Config)), default=None)



import random

import pygame
from pygame.rect import Rect

from .abstract_stim_maker import AbstractStimMaker
from .abstract_stim_maker import colors_dict


class RVvRHGVStimMaker(AbstractStimMaker):
    """Make visual search stimuli with rectangles,
    where target is red vertical rectangle. Half of distractors
    will be vertical green rectangles and the other half will
    be horizontal red rectangles. If the target is present,
    then (half - 1) of the distractors will be horizontal red
    rectangles."""

    def _make_stim(self,
                   xx_to_use_ctr,
                   yy_to_use_ctr,
                   target_inds,
                   cells_to_use,
                   display_surface):
        """helper function used by make_stim
        that sub-classes can override if they need to do something more
        complicated when making the visual search stimulus"""
        target_indices = []
        distractor_indices = []
        if self.grid_size:
            grid_as_char = [''] * self.num_cells
        else:
            grid_as_char = None

        # divide distractors up into target color-but-horizontal + distractor color-green vertical bars
        set_size = len(xx_to_use_ctr)
        num_distractors = set_size - len(target_inds)
        num_vert_rect = set_size // 2
        if len(target_inds) == 0:
            num_horz_rect = set_size // 2 - len(target_inds)
        else:
            num_horz_rect = set_size // 2

        if num_distractors % 2 == 1:  # e.g., if odd set size and target absent
            diff = num_distractors - (num_vert_rect + num_horz_rect)
            for _ in range(diff):
                if random.uniform(0, 1) > 0.5:
                    num_vert_rect += 1
                else:
                    num_horz_rect += 1

        distractor_orientation = list('V' * num_vert_rect + 'H' * num_horz_rect)
        random.shuffle(distractor_orientation)

        for item, (center_x, center_y) in enumerate(zip(xx_to_use_ctr, yy_to_use_ctr)):
            # notice we are now using PyGame order of sizes, (width, height)
            item_bbox_tuple = (0, 0) + (self.item_bbox_size[1], self.item_bbox_size[0])
            item_bbox = Rect(item_bbox_tuple)
            center = (int(center_x), int(center_y))
            item_bbox.center = center

            if item in target_inds:
                color = self.target_color
                rotate = False
                target_indices.append(center)
                if self.grid_size:
                    grid_as_char[cells_to_use[item]] = 't'
            else:
                orientation = distractor_orientation.pop()
                if orientation == 'V':
                    color = self.distractor_color
                    rotate = False
                elif orientation == 'H':
                    color = self.target_color
                    rotate = True

                distractor_indices.append(center)
                if self.grid_size:
                    grid_as_char[cells_to_use[item]] = orientation
                    # instead of 'd', so we know which distractor is which

            if type(color) == str:
                color = colors_dict[color]

            self.draw_item(display_surface=display_surface,
                           item_bbox=item_bbox,
                           color=color,
                           rotate=rotate)

        return grid_as_char, target_indices, distractor_indices

    def draw_item(self, display_surface, item_bbox, color, rotate):
        """Draws a vertical rectangle that is 1/3 the width of the item bounding box.

        Parameters
        ----------
        display_surface : pygame.Surface
            instance of Surface on which to draw item
        item_bbox : pygame.Rect
            instance of Rect that represents 'bounding box' within which
            item should be drawn.

        Returns
        -------
        None
        """
        # rect_to_draw is same object as item_bbox, but we're going to change the
        # "bounding box" into the items we will draw
        rect_to_draw = item_bbox

        if rotate:
            height = rect_to_draw.height // 3
            rect_to_draw.height = height
            rect_to_draw.bottom = rect_to_draw.bottom + height
        else:
            width = rect_to_draw.width // 3
            rect_to_draw.width = width
            rect_to_draw.left = rect_to_draw.left + width

        pygame.draw.rect(display_surface, color, rect_to_draw)



from .__about__ import (
    __author__,
    __commit__,
    __copyright__,
    __email__,
    __license__,
    __summary__,
    __title__,
    __uri__,
    __version__,
)

from .stim_makers import RVvGVStimMaker, Two_v_Five_StimMaker
from . import main
